@@ -3,14 +3,12 @@ from __future__ import print_function
 
 import io
 import os
+import platform
 import time
 from tempfile import TemporaryFile
+from unittest import mock
 
-try:
-    from unittest import mock
-except ImportError:
-    import mock
-
+import wurlitzer
 from wurlitzer import (
     PIPE,
     STDOUT,
@@ -138,8 +136,25 @@ def test_fd_leak():
 
 def test_buffer_full():
     with pipes(stdout=None, stderr=io.StringIO()) as (stdout, stderr):
-        long_string = "x" * 1000000  # create a very long string
+        long_string = "x" * 100000  # create a long string (longer than 65536)
         printf_err(long_string)
 
     # Test never reaches here as the process hangs.
     assert stderr.getvalue() == long_string + "\n"
+
+
+def test_buffer_full_default():
+    with pipes() as (stdout, stderr):
+        long_string = "x" * 100000  # create a long string (longer than 65536)
+        printf(long_string)
+
+    # Test never reaches here as the process hangs.
+    assert stdout.read() == long_string + "\n"
+
+
+def test_pipe_max_size():
+    max_pipe_size = wurlitzer._get_max_pipe_size()
+    if platform.system() == 'Linux':
+        assert 65535 <= max_pipe_size <= 1024 * 1024
+    else:
+        assert max_pipe_size is None
