@@ -4,7 +4,7 @@ Use `wurlitzer.pipes` or `wurlitzer.sys_pipes` as context managers.
 """
 from __future__ import print_function
 
-__version__ = '2.1.0'
+__version__ = '3.0.0'
 
 __all__ = [
     'pipes',
@@ -18,34 +18,76 @@ import ctypes
 import errno
 import io
 import os
-from contextlib import contextmanager
-from fcntl import F_GETFL, F_SETFL, fcntl
-
-try:
-    from queue import Queue
-except ImportError:
-    from Queue import Queue
-
-try:
-    import selectors
-except ImportError:
-    # py < 3.4
-    import selectors2 as selectors
-
+import platform
+import selectors
 import sys
 import threading
 import time
 import warnings
+from contextlib import contextmanager
+from fcntl import F_GETFL, F_SETFL, fcntl
+from functools import lru_cache
+from queue import Queue
+
+try:
+    from fcntl import F_SETPIPE_SZ
+except ImportError:
+    # ref: linux fcntl.h
+    F_SETPIPE_SZ = 1024 + 7
 
 libc = ctypes.CDLL(None)
 
+
+def _get_streams_cffi():
+    """Use CFFI to lookup stdout/stderr pointers
+
+    Should work ~everywhere, but requires compilation
+    """
+    try:
+        import cffi
+    except ImportError:
+        raise ImportError(
+            "Failed to lookup stdout symbols in libc. Fallback requires cffi."
+        )
+
+    try:
+        _ffi = cffi.FFI()
+        _ffi.cdef("const size_t c_stdout_p();")
+        _ffi.cdef("const size_t c_stderr_p();")
+        _lib = _ffi.verify(
+            '\n'.join(
+                [
+                    "#include <stdio.h>",
+                    "const size_t c_stdout_p() { return (size_t) (void*) stdout; }",
+                    "const size_t c_stderr_p() { return (size_t) (void*) stderr; }",
+                ]
+            )
+        )
+        c_stdout_p = ctypes.c_void_p(_lib.c_stdout_p())
+        c_stderr_p = ctypes.c_void_p(_lib.c_stderr_p())
+    except Exception as e:
+        warnings.warn(
+            "Failed to lookup stdout with cffi: {}.\nStreams may not be flushed.".format(
+                e
+            )
+        )
+        return (None, None)
+    else:
+        return c_stdout_p, c_stderr_p
+
+
+c_stdout_p = c_stderr_p = None
 try:
     c_stdout_p = ctypes.c_void_p.in_dll(libc, 'stdout')
     c_stderr_p = ctypes.c_void_p.in_dll(libc, 'stderr')
-except ValueError:  # pragma: no cover
-    # libc.stdout is has a funny name on OS X
-    c_stdout_p = ctypes.c_void_p.in_dll(libc, '__stdoutp')  # pragma: no cover
-    c_stderr_p = ctypes.c_void_p.in_dll(libc, '__stderrp')  # pragma: no cover
+except ValueError:
+    # libc.stdout has a funny name on macOS
+    try:
+        c_stdout_p = ctypes.c_void_p.in_dll(libc, '__stdoutp')
+        c_stderr_p = ctypes.c_void_p.in_dll(libc, '__stderrp')
+    except ValueError:
+        c_stdout_p, c_stderr_p = _get_streams_cffi()
+
 
 STDOUT = 2
 PIPE = 3
@@ -73,15 +115,56 @@ def dup2(a, b, timeout=3):
         raise dup_err
 
 
-class Wurlitzer(object):
+@lru_cache()
+def _get_max_pipe_size():
+    """Get max pipe size
+
+    Reads /proc/sys/fs/pipe-max-size on Linux.
+    Always returns None elsewhere.
+
+    Returns integer (up to 1MB),
+    or None if no value can be determined.
+
+    Adapted from wal-e, (c) 2018, WAL-E Contributors
+    used under BSD-3-clause
+    """
+    if platform.system() != 'Linux':
+        return
+
+    # If Linux procfs (or something that looks like it) exposes its
+    # maximum F_SETPIPE_SZ, adjust the default buffer sizes.
+    try:
+        with open('/proc/sys/fs/pipe-max-size', 'r') as f:
+            # Figure out OS max pipe size
+            pipe_max_size = int(f.read())
+    except Exception:
+        pass
+    else:
+        if pipe_max_size > 1024 * 1024:
+            # avoid unusually large values, limit to 1MB
+            return 1024 * 1024
+        elif pipe_max_size <= 65536:
+            # smaller than default, don't do anything
+            return None
+        else:
+            return pipe_max_size
+
+
+class Wurlitzer:
     """Class for Capturing Process-level FD output via dup2
 
-    Typically used via `wurlitzer.capture`
+    Typically used via `wurlitzer.pipes`
     """
 
     flush_interval = 0.2
 
-    def __init__(self, stdout=None, stderr=None, encoding=_default_encoding):
+    def __init__(
+        self,
+        stdout=None,
+        stderr=None,
+        encoding=_default_encoding,
+        bufsize=_get_max_pipe_size(),
+    ):
         """
         Parameters
         ----------
@@ -91,6 +174,10 @@ class Wurlitzer(object):
             The stream for forwarding stderr.
         encoding: str or None
             The encoding to use, if streams should be interpreted as text.
+        bufsize: int or None
+            Set pipe buffer size using fcntl F_SETPIPE_SZ (linux only)
+            default: use /proc/sys/fs/pipe-max-size up to a max of 1MB
+            if 0, will do nothing.
         """
         self._stdout = stdout
         if stderr == STDOUT:
@@ -98,6 +185,9 @@ class Wurlitzer(object):
         else:
             self._stderr = stderr
         self.encoding = encoding
+        if bufsize is None:
+            bufsize = _get_max_pipe_size()
+        self._bufsize = bufsize
         self._save_fds = {}
         self._real_fds = {}
         self._handlers = {}
@@ -114,6 +204,14 @@ class Wurlitzer(object):
         os.close(pipe_in)
         self._real_fds[name] = real_fd
 
+        # set max pipe buffer size (linux only)
+        if self._bufsize:
+            try:
+                fcntl(pipe_in, F_SETPIPE_SZ, self._bufsize)
+            except OSError:
+                # ignore failure to set pipe size
+                pass
+
         # make pipe_out non-blocking
         flags = fcntl(pipe_out, F_GETFL)
         fcntl(pipe_out, F_SETFL, flags | os.O_NONBLOCK)
@@ -151,8 +249,11 @@ class Wurlitzer(object):
         if self._stderr and sys.stderr:
             sys.stderr.flush()
 
-        libc.fflush(c_stdout_p)
-        libc.fflush(c_stderr_p)
+        if c_stdout_p is not None:
+            libc.fflush(c_stdout_p)
+
+        if c_stderr_p is not None:
+            libc.fflush(c_stderr_p)
 
     def __enter__(self):
         # flush anything out before starting
@@ -242,6 +343,7 @@ class Wurlitzer(object):
             flush_thread.join()
             # cleanup pipes
             [os.close(pipe) for pipe in pipes]
+            poller.close()
 
         self.thread = threading.Thread(target=forwarder)
         self.thread.daemon = True
@@ -268,11 +370,19 @@ class Wurlitzer(object):
 
 
 @contextmanager
-def pipes(stdout=PIPE, stderr=PIPE, encoding=_default_encoding):
+def pipes(stdout=PIPE, stderr=PIPE, encoding=_default_encoding, bufsize=None):
     """Capture C-level stdout/stderr in a context manager.
 
     The return value for the context manager is (stdout, stderr).
 
+    .. versionchanged:: 3.0
+
+        when using `PIPE` (default), the type of captured output
+        is `io.StringIO/BytesIO` instead of an OS pipe.
+        This eliminates max buffer size issues (and hang when output exceeds 65536 bytes),
+        but also means the buffer cannot be read with `.read()` methods
+        until after the context exits.
+
     Examples
     --------
 
@@ -281,14 +391,13 @@ def pipes(stdout=PIPE, stderr=PIPE, encoding=_default_encoding):
     ... output = stdout.read()
     """
     stdout_pipe = stderr_pipe = False
+    if encoding:
+        PipeIO = io.StringIO
+    else:
+        PipeIO = io.BytesIO
     # setup stdout
     if stdout == PIPE:
-        stdout_r, stdout_w = os.pipe()
-        stdout_w = os.fdopen(stdout_w, 'wb')
-        if encoding:
-            stdout_r = io.open(stdout_r, 'r', encoding=encoding)
-        else:
-            stdout_r = os.fdopen(stdout_r, 'rb')
+        stdout_r = stdout_w = PipeIO()
         stdout_pipe = True
     else:
         stdout_r = stdout_w = stdout
@@ -297,46 +406,39 @@ def pipes(stdout=PIPE, stderr=PIPE, encoding=_default_encoding):
         stderr_r = None
         stderr_w = stdout_w
     elif stderr == PIPE:
-        stderr_r, stderr_w = os.pipe()
-        stderr_w = os.fdopen(stderr_w, 'wb')
-        if encoding:
-            stderr_r = io.open(stderr_r, 'r', encoding=encoding)
-        else:
-            stderr_r = os.fdopen(stderr_r, 'rb')
+        stderr_r = stderr_w = PipeIO()
         stderr_pipe = True
     else:
         stderr_r = stderr_w = stderr
-    if stdout_pipe or stderr_pipe:
-        capture_encoding = None
-    else:
-        capture_encoding = encoding
-    w = Wurlitzer(stdout=stdout_w, stderr=stderr_w, encoding=capture_encoding)
+    w = Wurlitzer(stdout=stdout_w, stderr=stderr_w, encoding=encoding, bufsize=bufsize)
     try:
         with w:
             yield stdout_r, stderr_r
     finally:
         # close pipes
         if stdout_pipe:
-            stdout_w.close()
+            # seek to 0 so that it can be read after exit
+            stdout_r.seek(0)
         if stderr_pipe:
-            stderr_w.close()
+            # seek to 0 so that it can be read after exit
+            stderr_r.seek(0)
 
 
-def sys_pipes(encoding=_default_encoding):
+def sys_pipes(encoding=_default_encoding, bufsize=None):
     """Redirect C-level stdout/stderr to sys.stdout/stderr
 
     This is useful of sys.sdout/stderr are already being forwarded somewhere.
 
     DO NOT USE THIS if sys.stdout and sys.stderr are not already being forwarded.
     """
-    return pipes(sys.stdout, sys.stderr, encoding=encoding)
+    return pipes(sys.stdout, sys.stderr, encoding=encoding, bufsize=bufsize)
 
 
 _mighty_wurlitzer = None
 _mighty_lock = threading.Lock()
 
 
-def sys_pipes_forever(encoding=_default_encoding):
+def sys_pipes_forever(encoding=_default_encoding, bufsize=None):
     """Redirect all C output to sys.stdout/err
 
     This is not a context manager; it turns on C-forwarding permanently.
@@ -344,7 +446,7 @@ def sys_pipes_forever(encoding=_default_encoding):
     global _mighty_wurlitzer
     with _mighty_lock:
         if _mighty_wurlitzer is None:
-            _mighty_wurlitzer = sys_pipes(encoding)
+            _mighty_wurlitzer = sys_pipes(encoding, bufsize)
             _mighty_wurlitzer.__enter__()
 
 
